# Copyright (c) 2020 Slavfox
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Query builder.
"""
from typing import TYPE_CHECKING, Iterable, Iterator
from abc import ABC, abstractmethod

from snecs._filters import compile_filter, matches
from snecs.world import default_world

if TYPE_CHECKING:
    from typing import (
        Any,
        Tuple,
        Optional,
        Type,
        List,
        TypeVar,
        AbstractSet,
        Collection,
    )
    from snecs.component import Component
    from snecs.world import World
    from snecs.typedefs import EntityID
    from snecs._filters import Term, CompiledFilter

    QueryRow = Tuple[EntityID, List[Component]]
    QueryIterator = Iterator[QueryRow]
    DQ = TypeVar("DQ", bound="Query")

    T = TypeVar("T")

    def set_intersection(
        *s: "AbstractSet[EntityID]",  # noqa
    ) -> AbstractSet[EntityID]:
        ...


else:
    # get rid of the attribute lookup for performance
    set_intersection = set.intersection


__all__ = ["Query"]

_EMPTY_SET: "AbstractSet[EntityID]" = frozenset()


class BaseQuery(Iterable["QueryRow"], ABC):

    __slots__ = ("world", "component_types")

    def __init__(
        self,
        component_types: "Collection[Type[Component]]",
        world: "World" = default_world,
    ) -> "None":
        self.world = world
        self.component_types: "Collection[Type[Component]]" = component_types

    @abstractmethod
    def __iter__(self) -> "QueryIterator":
        ...


class CompiledQuery(BaseQuery, ABC):
    """
    Base class for compiled queries. See `query.compile`.

    Compiled queries are not filterable further. Filter them before
    compilation!
    """

    __slots__ = ()


class EmptyCompiledQuery(CompiledQuery):
    __slots__ = ()

    def __iter__(self) -> "QueryIterator":
        return
        yield  # type: ignore # noqa: W0101
        # The presence of a `yield` marks the function of a generator. This
        # is correct; a lone `return` would not work for defining an empty
        # generator.


class CompiledFilterQuery(CompiledQuery):
    """
    A compiled query with filters.

    The filters get compiled to a very fast expression; filtering a
    non-compiled, dynamic query is O(n), where ``n`` is the number of
    individual expressions in the filters (eg. ``A & B`` is three expressions
    - first matching against ``A``, then against ``B``, and then taking the
    ``and`` of the two results).

    A compiled filter, on the other hand, is O(n), but where (n) is the
    number of ``or`` expressions specifically. A compiled ``A & B & ~C & ...``
    takes the same amount of time (slightly less, even) to check as a
    dynamic ``A``.
    """

    __slots__ = ("_filter",)

    def __init__(
        self,
        component_types: "Collection[Type[Component]]",
        filter_: "CompiledFilter",
        world: "World" = default_world,
    ) -> "None":
        self._filter = filter_
        super().__init__(component_types, world)

    def __iter__(self) -> "QueryIterator":
        # The body of this method is mostly duplicated in every Query
        # subclass for performance.
        # Python, pls add macros. I want macros. Stringy codegen is a pain.
        entcache = self.world._entity_cache
        # list comprehensions are faster than generators if we don't need to
        # have an early exit.
        valid_entities: "AbstractSet[EntityID]" = set_intersection(
            *[entcache.get(ct, _EMPTY_SET) for ct in self.component_types]
        )
        # Moving attributes into the local scope to avoid doing attribute
        # lookups in a loop
        fits = self._filter.matches
        entmasks = self.world._entity_bitmasks
        entities = self.world._entities
        cmptypes = self.component_types
        for entid in valid_entities:
            if fits(entmasks[entid]):
                entcmps = entities[entid]
                yield entid, [entcmps[c] for c in cmptypes]


class CompiledRawQuery(CompiledQuery):
    __slots__ = ()

    def __iter__(self) -> "QueryIterator":
        entcache = self.world._entity_cache
        valid_entities: "AbstractSet[EntityID]" = set_intersection(
            *[entcache.get(ct, _EMPTY_SET) for ct in self.component_types]
        )
        entities = self.world._entities
        cmptypes = self.component_types
        for entid in valid_entities:
            entcmps = entities[entid]
            yield entid, [entcmps[c] for c in cmptypes]


class Query(BaseQuery):
    """
    An expression for querying the world for entities and Components.

    A query is instantiated by passing in a collection of component types to
    match, and is iterable, returning an iterator over tuples of
    ``(EntityId, List[Component])``.

    For instance, if you instantiate a query as::

        my_query = query((ComponentA, ComponentB))

    You can iterate over the results with::

        for entity_id, (component_a, component_b) in my_query:
            ...

    Note that the component types passed in at instantiation have to all be
    present on an entity, and will be included in the result. If you want to
    filter the query further, or want some of the components to not be
    returned, ``snecs`` overloads the three basic boolean operators on
    Component subclasses to let you filter queries in an expressive way::

        query(Position).filter((Velocity | Acceleration) & ~Frozen)

    The above query will return the entity_id and Position component for all
    entities that have a Position, are not Frozen, and have a Velocity or
    Acceleration (or both).
    """

    __slots__ = ("_filters",)

    def __new__(  # type: ignore
        cls,  # noqa: W0613
        component_types: "Collection[Type[Component]]",
        *args: "Any",
        **kwargs: "Any",
    ) -> "Query":
        """
        Instantiate a new query.

        If component_types is empty, return an EmptyQuery.

        :param component_types: Component types to query for.
        :type component_types: Collection[Type[Component]]
        """
        if not component_types:
            # Explicitly defer to object.__new__ to avoid a recursive call
            # to Query.__new__, which EmptyQuery inherits
            return object.__new__(EmptyQuery)  # type: ignore
        else:
            return super().__new__(cls)  # type: ignore

    def __init__(
        self,
        component_types: "Collection[Type[Component]]",
        world: "World" = default_world,
    ) -> "None":
        super().__init__(component_types, world)
        self._filters: "Optional[Term]" = None

    def filter(self: "DQ", expr: "Term") -> "DQ":
        """
        Filter this query according to a filter expression.
        """
        oldfilter = self._filters
        if oldfilter is None:
            self._filters = expr
        else:
            self._filters = oldfilter & expr
        return self

    def compile(self) -> "CompiledQuery":
        """
        Compile this query into a much faster one.

        This brings a big performance gain when using filters, and a
        negligible one (but still a performance gain, nevertheless) with
        unfiltered queries.

        Queries are intended to be compiled outside the hot path, like so::

            # compiled when the file is processed (imported)
            momentum_query = query((Position, Velocity), MY_WORLD).compile()

            def process_momentum():
                for entity, (Position, Velocity) in momentum_query:
                    ...
        """
        if not self.component_types:
            return EmptyCompiledQuery(self.component_types, world=self.world)
        elif self._filters is None:
            return CompiledRawQuery(self.component_types, world=self.world)
        return CompiledFilterQuery(
            self.component_types, compile_filter(self._filters), self.world
        )

    def __iter__(self) -> "QueryIterator":
        """
        Iterate over the results of this query.

        This does most of the heavy lifting before the actual loop, by taking
        an intersection of the sets of entities that have a specific
        component for each of the components we're querying for.
        """
        entcache = self.world._entity_cache
        cmptypes = self.component_types
        fst, *rest = cmptypes
        if rest:
            valid_entities: "AbstractSet[EntityID]" = set_intersection(
                *[entcache.get(ct, _EMPTY_SET) for ct in cmptypes]
            )
        else:
            # elide the set intersection if we don't need to do it
            valid_entities = entcache.get(fst, _EMPTY_SET)
        flt = self._filters
        entities = self.world._entities
        # If this query is filtered, match against each result individually.
        if flt is not None:
            entmasks = self.world._entity_bitmasks
            for entid in valid_entities:
                if matches(flt, entmasks[entid]):
                    entcmps = entities[entid]
                    yield entid, [entcmps[c] for c in cmptypes]
        # If not, just yield the results.
        else:
            for entid in valid_entities:
                entcmps = entities[entid]
                yield entid, [entcmps[c] for c in cmptypes]


class EmptyQuery(Query):
    """
    An empty query that will never return results.
    """

    __slots__ = ()

    def __iter__(self) -> "QueryIterator":
        return
        yield  # type: ignore # noqa: W0101

    def compile(self) -> "CompiledQuery":
        return EmptyCompiledQuery(self.component_types, self.world)



# Copyright (c) 2020 Slavfox
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Defines the World - the central object in snecs, which manages all the data.
"""
from typing import TYPE_CHECKING, cast

from snecs.component import Component

if TYPE_CHECKING:
    from typing import Type, Set, Dict, Optional
    from snecs._detail import Bitmask
    from snecs.typedefs import EntityID

__all__ = ["World", "default_world"]

DEFAULT_WORLD_NAME = "DEFAULT"


class World:
    """
    A container for all your entities, components and systems.

    Takes an optional ``name`` parameter that, if passed, will be used in the
    ``repr`` instead of the `id`.

    :param name: A human-readable name for this world, for debugging purposes.
                 The default world is called "default".
    :type name: str
    """

    _instances: "Dict[str, World]" = {}

    __slots__ = (
        "_entity_counter",
        "_entities",
        "_entity_bitmasks",
        "_entity_cache",
        "_entities_to_delete",
        "name",
    )

    def __init__(self, name: "Optional[str]" = None) -> None:
        if name:
            # We explicitly use World, and not self.__class__ here, to let
            # users still find World subclasses with World.get_by_name.
            if name in World._instances:
                raise ValueError(
                    f"A world with the name {name} already exists."
                )
            else:
                World._instances[name] = self

        self.name: "Optional[str]" = name
        self._entity_counter: "EntityID" = cast("EntityID", 0)
        self._entities: "Dict[EntityID, Dict[Type[Component], Component]]" = {}
        self._entity_bitmasks: "Dict[EntityID, Bitmask]" = {}
        self._entity_cache: "Dict[Type[Component], Set[EntityID]]" = {}
        self._entities_to_delete: "Set[EntityID]" = set()

    def __repr__(self) -> "str":
        """
        Return a pretty representation of the world, for debugging purposes.
        """
        if self.name is not None:
            return f"<{self.name} World>"
        else:
            return f"<World {id(self)} ({self._entity_counter} entities)>"

    @classmethod
    def get_by_name(cls, name: str) -> "World":
        """
        Get a World instance by name.

        This is obviously slower than just passing the World around, but it
        might sometimes be useful for avoiding circular imports. The
        preferred workflow is still to have one module that only imports
        from ``snecs`` and instantiates the world::

            from snecs import World

            my_world = World()

        And import ``my_world`` from there.

        :param name: Name of the `World` instance to get.
        :type name: str

        :return: The World with the given name, if it exists.
        :rtype: `World`

        :raises KeyError: If there is no World with the given name.
        """
        return World._instances[name]


#: The default World, used if you don't explicitly pass one.
default_world = World(DEFAULT_WORLD_NAME)



# Copyright (c) 2020 Slavfox
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Base Component classes and the functions necessary to make them work.
"""
from typing import TYPE_CHECKING
from abc import ABC, ABCMeta

from snecs._detail import Bitmask, InvariantDict
from snecs._filters import AndExpr, Expr, NotExpr, OrExpr

if TYPE_CHECKING:
    from typing import Type, Any, Dict, TypeVar
    from snecs._filters import Term

    CType = TypeVar("CType", bound="Component")

__all__ = ["Component", "RegisteredComponent", "register_component"]


class ComponentMeta(ABCMeta):
    """
    Metaclass for Components, supporting filter expression syntax.

    This exists because the magic methods have to be defined on the metaclass,
    which Component classes are instances of; just like defining ``__add__``
    on a class ``Foo`` lets you do::

        foo = Foo()
        foo + 1  # calls ``foo.__add__(1)``
        Foo + 1  # TypeError: ...

    Defining bitwise operations on the metaclass lets you apply them to
    *classes* with that metaclass::

        FooComponent & BarComponent  # calls ComponentMeta.__and__
        foo = FooComponent()
        bar = BarComponent()
        foo & bar  # TypeError: ...
    """

    _bitmask: "Bitmask"

    def __and__(self, other: "Term") -> "Term":
        if other is self:
            return self
        elif isinstance(other, Expr):
            return other.__rand__(self)
        return AndExpr(self, other)

    def __or__(self, other: "Term") -> "Term":
        if other is self:
            return self
        elif isinstance(other, Expr):
            return other.__ror__(self)
        return OrExpr(self, other)

    def __invert__(self) -> "NotExpr":
        return NotExpr(self)


class Component(ABC, metaclass=ComponentMeta):
    """
    Base Component class.

    snecs components *must* subclass `Component`, and be registered with
    `register_component` if they are going to be instantiated::

        @register_component
        class MyComponent(Component):
            ...

    Base and abstract Component classes don't have to be registered, as long as
    you're not going to instantiate them.

    Other than that, there are no restrictions on what Components should be.
    They can be anything as simple as an integer::

        @register_component
        class IntComponent(int, Component):
            pass

    Or as complex as you want::

        @register_component
        class BigComponent(Component):
            __slots__ = ('a', 'b', 'c', 'd', ...)

            def __init__(a, b, c, d, ...):
                ...

            def __str__(self):
                ...

            ...

    If you want to use the snecs full-world serialization, you should
    override `serialize` and `deserialize` in *all* your registered
    Components::

        @register_component
        class IntPairComponent(Component):
            ...
            def serialize(self):
                return (self.first, self.second)

            @classmethod
            def deserialize(cls, serialized):
                return cls(*serialized)

    You *must* define either both, or neither of those two methods.
    Registering a class that only defines one of those is an error.

    If you register a class that defines `serialize` or
    `deserialize`, *all* of your registered classes must define them.
    Likewise, if you register a class that doesn't define `serialize` and
    `deserialize`, *none* of your registered classes can define them.
    """

    _bitmask: "Bitmask"
    __slots__ = ()

    def serialize(self) -> "Any":  # type: ignore
        """
        Serialize an instance of this component into a simpler type.

        Override this in all your Component classes to make use of snecs'
        full-World serialization feature.
        """
        raise AttributeError(
            f"{self.__class__} doesn't define serialize()"
            f"and so is not serializable."
        )

    @classmethod
    def deserialize(  # type: ignore
        cls: "Type[CType]", serialized: "Any"
    ) -> "CType":
        """
        Deserialize a serialized instance of this component.

        Will get the output of `serialize` as an argument.

        Override this in all your Component classes to make use of snecs'
        full-World serialization feature.
        """
        raise AttributeError(
            f"{cls} doesn't define deserialize()"
            f"and so is not deserializable."
        )


class _ComponentRegistry(InvariantDict["Type[Component]", "Bitmask"]):
    __slots__ = ()

    def value_for(self, key: "Type[Component]") -> "Bitmask":
        return Bitmask(1 << len(self))


_component_registry: "_ComponentRegistry" = _ComponentRegistry()
_component_names: "Dict[str, Type[Component]]" = {}


def _overrides_serialize(cls: "Type[Component]") -> "bool":
    return cls.serialize is not Component.serialize  # type: ignore


def _overrides_deserialize(cls: "Type[Component]") -> "bool":
    # This is a bit of jank, so it deserves an explanation.
    #
    # serialize() is a normal method, so we can just test if it's the base
    # one directly. On the other hand, deserialize() is a class method - so
    # trying to access cls.deserialize goes through the classmethod
    # descriptor and returns a method bound to `cls`. This means it
    # will never be Component.deserialize. Uh oh!
    #
    # Luckily, we have another trick up our sleeve. Through that same
    # mechanism, B.deserialize is a bound *method*, and not a function (as
    # unbound methods are).
    #
    # You might remember that bound methods hold a reference to the
    # underlying function object in their .__func__ attribute. So, to check
    # if the class passed as an argument overrides `deserialize`, we first
    # check:
    if (
        getattr(cls.deserialize, "__func__", None)  # type: ignore
        is Component.deserialize.__func__  # type: ignore
    ):
        # and if so,
        return False
    # Otherwise, if we landed here, we know that `deserialize` is not a
    # classmethod, *or* it's a classmethod different from Component.serialize.
    # Either way, it's not the base implementation, so we happily
    return True


def register_component(component: "Type[CType]") -> "Type[CType]":
    """
    A decorator to register a class for use as a snecs component.

    Every class intended to be instantiated and used as a snecs component
    *must* inherit from `Component` and be registered by decorating it with
    ``@register_component``.
    """
    # we're going to be using this a lot
    cn = component.__name__
    if component in _component_registry:
        raise ValueError(
            f"Component class {cn} is already registered. Cannot register "
            f"the same component class twice. (did you accidentally "
            f"@register_component a RegisteredComponent subclass?)"
        )
    has_deserialize = _overrides_deserialize(component)
    has_serialize = _overrides_serialize(component)

    serializable = False

    if has_serialize:
        if not has_deserialize:
            raise TypeError(
                f"Component class {cn} has an overriden `serialize()` but "
                f"does not override `deserialize()`. You must implement "
                f"either both or neither of those methods in registered "
                f"component classes."
            )
        serializable = True
    elif has_deserialize:
        raise TypeError(
            f"Component class {cn} has an overriden `deserialize()` but does "
            f"not override `serialize()`. You must implement either both or "
            f"neither of those methods in registered component classes."
        )
    if cn in _component_names and serializable:
        raise ValueError(
            f"A Component class named {cn} is already registered. Cannot "
            f"register a serializable Component class with a non-unique name."
        )
    bitmask = _component_registry.add(component)
    _component_names[cn] = component
    component._bitmask = bitmask
    return component


class RegisteredComponent(Component):
    """
    A convenience `Component` subclass for auto-registering components.

    The following two ways of defining a component are equivalent::

        @register_component
        class MyComponent1(snecs.Component):
            ...

        class MyComponent2(snecs.RegisteredComponent):
            ...

    Bear in mind, however, that all subclasses of a RegisteredComponent
    subclass will also get registered. Registering unnecessary components
    does have a slight performance and memory impact, so it's better to
    explicitly use `register_component`, or use ``RegisteredComponent`` as a
    mixin on leaf classes only::

        class AbstractComponent(snecs.Component, ABC):
            ...

        class ConcreteComponent(AbstractComponent, RegisteredComponent):
            ...

    This last approach isn't much different than just decorating your
    components with `register_component`, though.
    """

    def __init_subclass__(cls: "Type[RegisteredComponent]") -> "None":
        """
        Automagically call `register_component` on subclasses.
        """
        super().__init_subclass__()
        register_component(cls)



# Copyright (c) 2020 Slavfox
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
snecs - a straightforward, nimble ECS for Python.

snecs is a pure Python 3.6+, dependency-free ECS library, heavily inspired by
Rust’s Legion, and aiming to be as fast and easy-to-use as possible.
"""
from snecs.component import Component, RegisteredComponent, register_component
from snecs.ecs import (
    add_component,
    add_components,
    all_components,
    delete_entity_immediately,
    deserialize_world,
    entity_component,
    entity_components,
    exists,
    has_component,
    has_components,
    new_entity,
    process_pending_deletions,
    remove_component,
    schedule_for_deletion,
    serialize_world,
)
from snecs.query import Query
from snecs.world import World

__all__ = [
    "World",
    "Component",
    "RegisteredComponent",
    "register_component",
    # Functions to interact with the actual ECS
    "new_entity",
    "add_component",
    "add_components",
    "entity_component",
    "entity_components",
    "has_component",
    "has_components",
    "all_components",
    "remove_component",
    "schedule_for_deletion",
    "delete_entity_immediately",
    "process_pending_deletions",
    "serialize_world",
    "deserialize_world",
    "exists",
    "Query",
]

__version_info__ = ("1", "1", "1")
__version__ = ".".join(__version_info__)



"""
Query filtering tools.

Effectively a boolean algebra system on bitvectors.
"""
from typing import TYPE_CHECKING, Callable
from abc import ABC, abstractmethod
from functools import reduce
from operator import and_, or_

from snecs._detail import ZERO, Bitmask

if TYPE_CHECKING:
    from typing import (
        Tuple,
        Sequence,
        ClassVar,
        TypeVar,
        List,
        Union,
        NoReturn,
    )

    from snecs.component import ComponentMeta

    _Selector = Bitmask
    _ValueMask = Bitmask
    _MatcherCombinerType = Callable[
        ["ExprCompiler", "ExprCompiler"], "ExprCompiler"
    ]
    _BoolLike = Union["int", "bool"]
    _CompiledMatcherType = Callable[[Bitmask], _BoolLike]

    T = TypeVar("T")
    Term = Union["Expr", "ComponentMeta"]

    # make inv pass type checking
    def inv(term: "Term") -> "Term":  # noqa
        ...


else:
    from operator import inv


__all__ = ["compile_filter"]


class NonliteralTerm(ABC):
    __slots__ = ()

    @abstractmethod
    def matches(self, bitmask: "Bitmask") -> "_BoolLike":
        ...


class CompiledFilter(NonliteralTerm, ABC):
    __slots__ = ("_expr",)

    def __init__(self, ex: "str") -> "None":
        self._expr = ex

    def __repr__(self) -> "str":
        return f"CompiledFilter[{self._expr}]"


def compile_filter(term: "Term") -> "CompiledFilter":
    """
    Compile a filter into a much faster function (Bitmask) -> `bool`.

    :param term: Filter expression to compile
    :return: a function taking a Bitmask and returning `True` if it
             matches the filter and `False` if it doesn't.
    """
    if isinstance(term, Expr):
        return term._make_compiler().compile(str(term))
    # else
    term_bitmask = term._bitmask

    # This is ugly and all, but we're doing this the ugly way for maximum
    # performance; staticmethod is a descriptor, so it adds a little bit of
    # overhead to the name lookup, so calling filter.matches() is slower
    # than normal methods. OTOH, when saving it to an intermediate variable:
    #
    #     matches = filter.matches
    #     for ... in ...:
    #         if matches(...): ...
    #
    # staticmethod and an external function are about on par, with normal
    # methods being slowest. Since the external function is the fastest in
    # both cases, that's what we use - but we define an unused first argument
    # to keep the signature correct.
    def match(_: object, bitmask: "Bitmask") -> "_BoolLike":
        # calling bool() on this to hit the fast-path for booleans is not
        # worth the function call overhead.
        return bitmask & term_bitmask

    class Filter(CompiledFilter):
        __slots__ = ()
        matches: "_CompiledMatcherType" = match  # type: ignore

    return Filter(term.__name__)


def matches(
    term: "Union[Term, CompiledFilter]", bitmask: "Bitmask"
) -> "_BoolLike":
    """
    Check if a filter expression matches a bitmask.

    :param term: The expression to match the bitmask against
    :param bitmask: Bitmask to match
    """
    if isinstance(term, NonliteralTerm):
        return term.matches(bitmask)
    else:
        return bitmask & term._bitmask


def _format_expr_term(term: "Term") -> str:
    if isinstance(term, Expr):
        return f"({str(term)})"
    else:
        return term.__name__


class ExprCompiler:
    """
    A compiler for an expression tree.

    ExprCompiler represents an arbitrary expression tree as a flat list of
    pairs of selector bitmasks and value masks. For example, for the
    bitmasks::

        A: 0001
        B: 0010
        C: 0100

    And the expression::

        A & (B | C)

    The build process happens as follows:


    1. We make an ExprCompiler for ``A``, with the selector and value mask
       both equal to ``0001``.
    2. We recurse into the second operand of the AND, ``(B | C)``, and:
        1. We make an ExprCompiler for ``B``, with the selector and value mask
           both equal to ``0010``.
        2. We make an ExprCompiler for ``C``, with the selector and value mask
           both equal to ``0100``.
        3. We OR the two ExprCompiler together, returning a new ExprCompiler
           with the two as separate clauses::

               ExprCompiler(clauses=[(0010, 0010), (0100, 0100)])
        4. We return the new ExprCompiler up to the caller
    3. We AND the ``A`` ExprCompiler with the ``(B | C)`` ExprCompiler, by
       ORing the bitmasks of each clause with ``A``::

           0010 | 0001 = 0011
           0100 | 0001 = 0101



           result = ExprCompiler(clauses=[(0011, 0011), (0101, 0101)])
    4. We compile that into a function of the form::

           clauses = [(0011, 0011), (0101, 0101)]
           def match(bitmask):
               return any(
                   bitmask & selector == valuemask
                   for selector, valuemask in clauses
               )
    """

    __slots__ = ("clauses",)

    def __init__(
        self, clauses: "Sequence[Tuple[_Selector, _ValueMask]]"
    ) -> None:
        # Sanity check to trim out ignored bits, so that eg.
        #
        #   selector:   0b_0110
        #   value_mask: 0b_1111
        #
        # doesn't break `bitmask & selector == valuemask`
        self.clauses = [
            (selector, selector & valuemask) for selector, valuemask in clauses
        ]

    def __and__(self, other: "ExprCompiler") -> "ExprCompiler":
        new_clauses = []
        for lselector, lvaluemask in self.clauses:
            for rselector, rvaluemask in other.clauses:
                common_selector = lselector & rselector
                if common_selector:
                    # if (x & ~x), drop the clause
                    if lvaluemask & common_selector != rvaluemask & rselector:
                        continue
                new_clauses.append(
                    ((lselector | rselector), (lvaluemask | rvaluemask))
                )
        return ExprCompiler(new_clauses)

    def __or__(self, other: "ExprCompiler") -> "ExprCompiler":
        return ExprCompiler([*self.clauses, *other.clauses])

    def _pretty_clauses(self) -> "List[str]":
        if not self.clauses:
            return []
        max_length = max(len(f"{clause[0]:b}") for clause in self.clauses)
        pretty_clauses = []
        for selector, valuemask in self.clauses:
            valuemask_str = f"{valuemask:b}".zfill(max_length)
            pretty_selector = f"{selector:b}".zfill(max_length).replace(
                "0", "."
            )
            pretty_clauses.append(
                "".join(
                    valuemask_str[i] if ch != "." else ch
                    for i, ch in enumerate(pretty_selector)
                )
            )
        return pretty_clauses

    def compile(
        self, expr_repr: "str" = "<unknown expression>"
    ) -> "CompiledFilter":
        clauses = self.clauses
        if not self.clauses:
            # W0613 = "Unused argument"
            # FalseExpr
            def match(_: object, bitmask: "Bitmask") -> bool:  # noqa: W0613
                return False

        elif len(self.clauses) == 1 and self.clauses[0][0] == 0:
            # TrueExpr
            def match(_: object, bitmask: "Bitmask") -> bool:  # noqa: W0613
                return True

        elif len(self.clauses) == 1:
            # Fast path for single-clauses
            selector, valuemask = self.clauses[0]

            def match(_: object, bitmask: "Bitmask") -> bool:
                return bitmask & selector == valuemask

        else:

            def match(_: object, bitmask: "Bitmask") -> bool:
                return any(
                    bitmask & selector == valuemask
                    for selector, valuemask in clauses
                )

        class Filter(CompiledFilter):
            matches: "_CompiledMatcherType" = match  # type: ignore

        return Filter(expr_repr)


# Expression classes


class Expr(NonliteralTerm, ABC):
    """
    Base class for boolean expressions.
    """

    __slots__ = ()

    @abstractmethod
    def _make_compiler(self) -> "ExprCompiler":
        """
        Build an ExprCompiler out of this Expr's expression tree.
        """
        ...

    @abstractmethod
    def __invert__(self) -> "Term":
        ...

    @abstractmethod
    def __and__(self, other: "Term") -> "Term":
        ...

    @abstractmethod
    def __rand__(self, other: "Term") -> "Term":
        ...

    @abstractmethod
    def __or__(self, other: "Term") -> "Term":
        ...

    @abstractmethod
    def __ror__(self, other: "Term") -> "Term":
        ...

    @abstractmethod
    def matches(self, bitmask: "Bitmask") -> "bool":
        ...

    @abstractmethod
    def __str__(self) -> "str":
        ...

    def __repr__(self) -> "str":
        return f"{self.__class__}[{str(self)}]"


class StaticExpr(Expr, ABC):
    """
    Base class for expressions reducible to a form with no variables.
    """

    __slots__ = ()
    _retval: "bool" = False

    def __str__(self) -> "str":
        return str(self._retval).lower()

    def matches(self, bitmask: "Bitmask") -> "bool":
        return self._retval

    def __eq__(self, other: "object") -> "bool":
        if isinstance(other, self.__class__):
            return True
        return False

    def __hash__(self) -> "int":
        return hash(self.__class__)


class _TrueExpr(StaticExpr):
    """
    The simplified form of a expression that is always true, like A | ~A.

    Matches everything.
    """

    __slots__ = ()
    _retval = True

    def __invert__(self) -> "_FalseExpr":
        return FalseExpr

    def _make_compiler(self) -> "ExprCompiler":
        # Anything & 0 == 0
        return ExprCompiler(clauses=[(Bitmask(0), Bitmask(0))])

    def __or__(self, other: "Term") -> "_TrueExpr":
        return self

    __ror__ = __or__

    def __and__(self, other: "Term") -> "Term":
        return other

    __rand__ = __and__


class _FalseExpr(StaticExpr):
    """
    The simplified form of an unsatisfiable expression, like A & ~A.

    Matches nothing.
    """

    __slots__ = ()
    _retval = False

    def __invert__(self) -> "_TrueExpr":
        return TrueExpr

    def _make_compiler(self) -> "ExprCompiler":
        # any([]) == False
        return ExprCompiler(clauses=[])

    def __or__(self, other: "Term") -> "Term":
        return other

    __ror__ = __or__

    def __and__(self, other: "Term") -> "_FalseExpr":
        return self

    __rand__ = __and__


TrueExpr = _TrueExpr()
FalseExpr = _FalseExpr()


class DynExpr(Expr, ABC):
    """
    Base abstract class for "dynamic" filter expressions, that have arguments.

    Represents an unary expression, with a single operand.
    """

    __slots__ = ("matcher", "terms")
    _operator_repr: "ClassVar[str]" = "(Expr)"
    _matcher_combiner: "ClassVar[_MatcherCombinerType]" = and_

    @abstractmethod
    def _make_compiler(self) -> "ExprCompiler":
        ...

    def __init__(self, *terms: "Term"):
        self.terms: "Tuple[Term, ...]" = terms

    def __and__(self, other: "Term") -> "Term":
        if isinstance(other, AndExpr):
            return other.__rand__(self)
        return AndExpr(self, other)

    def __rand__(self, other: "Term") -> "Term":
        if isinstance(other, AndExpr):
            return other & self
        return AndExpr(other, self)

    def __or__(self, other: "Term") -> "Term":
        if isinstance(other, OrExpr):
            return other.__ror__(self)
        return OrExpr(self, other)

    def __ror__(self, other: "Term") -> "Term":
        if isinstance(other, OrExpr):
            return other | self
        return OrExpr(self, other)

    def __invert__(self) -> "Term":
        return NotExpr(self)

    def _make_compiler_for(self, term: "Term") -> "ExprCompiler":
        if isinstance(term, Expr):
            return self._extend_compiler(term._make_compiler())
        else:
            return self._get_component_compiler(term)

    def _get_component_compiler(self, term: "ComponentMeta") -> "ExprCompiler":
        return ExprCompiler(clauses=[(term._bitmask, term._bitmask)])

    def _extend_compiler(self, matcher: "ExprCompiler") -> "ExprCompiler":
        return matcher

    def __hash__(self) -> int:
        return hash((self.__class__, self.terms))

    def __eq__(self, other: "object") -> bool:
        return isinstance(other, self.__class__) and set(other.terms) == set(
            self.terms
        )


class NotExpr(DynExpr):
    """
    A "not" expression: `~Term`.

    Matches bitmasks where all the bits from ``Term._bitmask`` are zero.
    """

    __slots__ = ()
    _operator_repr = "~"

    def __init__(self, term: "Term") -> "None":  # noqa: W0235
        # W0235: "Useless super delegation in method '__init__'"
        # Actually not useless, because NotExpr.__init__ is not variadic.
        super().__init__(term)

    def matches(self, bitmask: "Bitmask") -> "bool":
        return not matches(self.terms[0], bitmask)

    def _make_compiler(self) -> "ExprCompiler":
        return self._make_compiler_for(self.terms[0])

    def _get_component_compiler(self, term: "ComponentMeta") -> ExprCompiler:
        return ExprCompiler(clauses=[(term._bitmask, ZERO)])

    def _extend_compiler(
        self, matcher: "ExprCompiler"
    ) -> "NoReturn":  # pragma: no cover
        raise TypeError("NotExpr._extend_compiler intentionally omitted.")

    def __invert__(self) -> "Term":
        return self.terms[0]

    def __or__(self, other: "Term") -> "Term":
        if other == self.terms[0]:
            return TrueExpr
        return super().__or__(other)

    def __ror__(self, other: "Term") -> "Term":
        if other == self.terms[0]:
            return TrueExpr
        return super().__ror__(other)

    def __and__(self, other: "Term") -> "Term":
        if other == ~self:
            return FalseExpr
        return super().__and__(other)

    def __rand__(self, other: "Term") -> "Term":
        if other == ~self:
            return FalseExpr
        return super().__rand__(other)

    def __str__(self) -> "str":
        return f"~{_format_expr_term(self.terms[0])}"


class MultiExpr(DynExpr, ABC):
    """
    An abstract base class for 2+-ary expressions, like ``X & Y & Z``.
    """

    __slots__ = ()
    _operator_repr: "ClassVar[str]" = "(MultiExpr)"
    _matcher_combiner: "ClassVar[_MatcherCombinerType]" = and_

    def __str__(self) -> "str":
        return f"""{
            self._operator_repr.join(_format_expr_term(t) for t in self.terms)
            }"""

    def _make_compiler(self) -> "ExprCompiler":
        return reduce(
            # see: https://github.com/python/mypy/issues/5485
            self._matcher_combiner,  # type: ignore
            (self._make_compiler_for(term) for term in self.terms),
        )


class AndExpr(MultiExpr):
    """
    An "and" expression: ``X & Y``.

    Matches bitmasks matched by all of its operands.
    """

    __slots__ = ()
    _operator_repr = " & "

    def __and__(self, other: "Term") -> "Term":
        if isinstance(other, AndExpr):
            terms = (*self.terms, *other.terms)
            first = terms[0]
            if all(term == first for term in terms):
                return first
            return AndExpr(*terms)
        if other not in self.terms:
            return AndExpr(*self.terms, other)
        return self

    def __rand__(self, other: "Term") -> "Term":
        if isinstance(other, AndExpr):
            terms = (*other.terms, *self.terms)
            first = terms[0]
            if all(term == first for term in terms):
                return first
            return AndExpr(*other.terms, *self.terms)
        if other not in self.terms:
            return AndExpr(other, *self.terms)
        return self

    def __invert__(self) -> "OrExpr":
        # De Morgan's first law
        # ¬(A ∧ B) ⇔ ¬A ∨ ¬B
        return OrExpr(*map(inv, self.terms))

    def matches(self, bitmask: "Bitmask") -> "bool":
        return all(matches(term, bitmask) for term in self.terms)


class OrExpr(MultiExpr):
    """
    An "or" expression: ``X & Y``.

    Matches bitmasks matched by any of its operands.
    """

    __slots__ = ()
    _operator_repr = " | "
    _matcher_combiner: "ClassVar[_MatcherCombinerType]" = or_

    def __or__(self, other: "Term") -> "OrExpr":
        if isinstance(other, OrExpr):
            return OrExpr(
                *self.terms,
                *[term for term in other.terms if term not in self.terms],
            )
        elif other not in self.terms:
            return OrExpr(*self.terms, other)
        return self

    def __ror__(self, other: "Term") -> "OrExpr":
        if isinstance(other, AndExpr):
            return OrExpr(
                *[term for term in other.terms if term not in self.terms],
                *self.terms,
            )
        elif other not in self.terms:
            return OrExpr(other, *self.terms)
        return self

    def __invert__(self) -> "AndExpr":
        # De Morgan's second law
        # ¬(A ∨ B) ⇔ ¬A ∧ ¬B
        return AndExpr(*map(inv, self.terms))

    def matches(self, bitmask: "Bitmask") -> "bool":
        return any(matches(term, bitmask) for term in self.terms)



# Copyright (c) 2020 Slavfox
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Type aliases to facilitate easier typing of snecs-dependent code.
"""
from typing import TYPE_CHECKING, Any, Dict

from snecs._filters import Expr as FilterExpressionType
from snecs.query import BaseQuery as QueryType
from snecs.query import CompiledQuery as CompiledQueryType

if TYPE_CHECKING:

    class EntityID(int):
        """
        A ``NewType(int)`` that only allows incrementation.
        """

        __slots__ = ()

        def __add__(self, other: "int") -> "EntityID":  # noqa D105
            ...

        def __iadd__(self, other: "int") -> "EntityID":  # noqa D105
            ...


else:
    #: Effectively an integer `typing.NewType` that only allows incrementation.
    EntityID = int


__all__ = [
    "FilterExpressionType",
    "QueryType",
    "CompiledQueryType",
    "EntityID",
    "SerializedWorldType",
]

SerializedWorldType = Dict[str, Any]  # type: ignore



# Copyright (c) 2020 Slavfox
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Functions for interacting with the ECS.

All the functions in this module are reexported under the ``snecs`` root for
convenience, so that you can::

    from snecs import new_entity

    new_entity(...)

    # or

    import snecs

    snecs.new_entity(...)
    snecs.add_component(...)

Instead of having to type ``snecs.ecs`` every time.
"""
from typing import TYPE_CHECKING
from typing import cast as _cast
from types import MappingProxyType as _MappingProxy

from snecs._detail import ZERO as _ZERO
from snecs.component import _component_names
from snecs.world import World as _World
from snecs.world import default_world as _default_world

if TYPE_CHECKING:
    from typing import (
        Type,
        Iterable,
        Collection,
        TypeVar,
        Mapping,
        Dict,
        Any,
        List,
        Optional,
    )
    from snecs.component import Component
    from snecs.typedefs import EntityID, SerializedWorldType

    C = TypeVar("C", bound=Component)

__all__ = [
    "add_component",
    "add_components",
    "all_components",
    "delete_entity_immediately",
    "deserialize_world",
    "entity_component",
    "entity_components",
    "exists",
    "has_component",
    "has_components",
    "new_entity",
    "process_pending_deletions",
    "remove_component",
    "schedule_for_deletion",
    "serialize_world",
    "SERIALIZED_COMPONENTS_KEY",
    "SERIALIZED_ENTITIES_KEY",
]


def new_entity(
    components: "Collection[Component]" = (), world: "_World" = _default_world
) -> "EntityID":
    """
    Create an entity in a World with the given Components, returning its ID.

    :param world: World to create the entity in.
    :type world: Optional[`World`]

    :param components: An iterable of Component instances to attach to the
                       entity.
    :type components: Collection[:class:`~snecs.Component`]

    :return: ID of the newly created entity.
    :rtype: `EntityID`
    """
    unique_ctypes = {c.__class__ for c in components}
    if len(unique_ctypes) < len(components):
        raise ValueError(
            f"Cannot create entity with components: {tuple(components)}. "
            f"Adding multiple components of the same type to one "
            f"entity is not allowed."
        )

    bitmask = _ZERO
    entdict = {}
    entcache = world._entity_cache
    id_ = world._entity_counter + 1
    for component in components:
        cc = component.__class__
        bitmask |= component._bitmask
        entdict[cc] = component
        entcache.setdefault(cc, set()).add(id_)

    world._entities[id_] = entdict
    world._entity_bitmasks[id_] = bitmask
    world._entity_counter = id_
    return id_


def add_component(
    entity_id: "EntityID",
    component: "Component",
    world: "_World" = _default_world,
) -> "None":
    """
    Add a single component instance to an entity.

    .. warning::

        Adding multiple components of the same type to a single entity is
        invalid.

    :param entity_id: ID of the entity to add the Component to.
    :type entity_id: `EntityID`

    :param component: A single Component instances to add to the Entity.
    :type component: `Component`

    :param world: The World holding the entity.
    :type world: Optional[`World`]

    :raises KeyError: If the Component type was not registered, or if the
                      entity doesn't exist in the given World.

    :raises ValueError: If the entity already has a Component of the same type
                        as the given one.
    """
    entd = world._entities[entity_id]
    cc = component.__class__
    if cc in entd:
        # Adding multiple components of the same time will break shit. Heavily.
        #
        # First of all, all the components of that type other than the
        # last-added one will be effectively inaccessible, because component
        # mappings and the query caches are keyed on component types.
        #
        # Worse than that, they will also be unremovable - since after
        # removing a component type from an entity, it's considered to not
        # longer have a component of that type.
        #
        # Instead of dealing with all that, we just explicitly disallow this.
        # This was an AssertionError previously, but I feel it's important
        # enough to be a proper runtime check.
        raise ValueError(
            f"Tried to add a {cc.__name__} component to entity {entity_id}, "
            f"but it already has a component of that type."
        )
    entd[cc] = component
    world._entity_cache.setdefault(cc, set()).add(entity_id)
    world._entity_bitmasks[entity_id] |= component._bitmask


def add_components(
    entity_id: "EntityID",
    components: "Collection[Component]",
    world: "_World" = _default_world,
) -> "None":
    """
    Add components to an entity.

    See notes in `add_component`.

    :param entity_id: ID of the entity to add the Components to.
    :type entity_id: `EntityID`

    :param components: An iterable of Component instances to add to the Entity.
    :type components: Collection[:class:`~snecs.Component`]

    :param world: The World holding the entity.
    :type world: Optional[`World`]

    :raises KeyError: If any of the Component types was not registered, or if
                      the entity doesn't exist in the given World.

    :raises ValueError: If the entity already has a Component of the same type
                        as one of the given ones.
    """
    entd = world._entities[entity_id]
    new_bitmask = _ZERO
    entcache = world._entity_cache
    unique_ctypes = {c.__class__ for c in components}
    if len(unique_ctypes) < len(components) or unique_ctypes & entd.keys():
        # This is an error, so doesn't have to be fast
        existing = [c.__class__.__name__ for c in entd.keys()]
        raise ValueError(
            f"Tried to add duplicate components for entity {entity_id}: "
            f"{components} (entity already has {existing})."
        )

    for c in components:
        cc = c.__class__
        new_bitmask |= c._bitmask
        # Waaay faster than entity.update({cc: c for ...}).
        entd[cc] = c
        entcache.setdefault(cc, set()).add(entity_id)
    world._entity_bitmasks[entity_id] = new_bitmask


def entity_component(
    entity_id: "EntityID",
    component_type: "Type[C]",
    world: "_World" = _default_world,
) -> "C":
    """
    Get the Component instance of a given type for a specific entity.

    :param entity_id: ID of the entity to get the Component from.
    :type entity_id: `EntityID`

    :param component_type: Type of the component to fetch.
    :type component_type: Type[:class:`~snecs.Component`]

    :param world: The World to look up the entity in.
    :type world: Optional[`World`]

    :return: The instance of ``component_type`` that's assigned to the entity.
    :rtype: Instance of ``component_type``

    :raises KeyError: If the entity doesn't exist in this `World` or if it
                      doesn't have the requested Component.
    """
    return world._entities[entity_id][component_type]  # type: ignore


def entity_components(
    entity_id: "EntityID",
    components: "Iterable[Type[Component]]",
    world: "_World" = _default_world,
) -> "Dict[Type[Component], Component]":
    """
    Get the given instances of given Component types for a specific entity.

    :param entity_id: ID of the entity to get the Components from.
    :type entity_id: `EntityID`

    :param components: An iterable of Component types to look up.
    :type components: Iterable[Type[:class:`~snecs.Component`]]

    :param world: The World to look up the entity in.
    :type world: Optional[`World`]

    :return: A dictionary mapping each given Component type to the instance of
             that type attached to the given entity. Note that mutations of
             this dictionary do not affect the world.
    :rtype: Dict[Type[Component], Component]

    :raises KeyError: If the entity doesn't exist in this `World` or if it
                      doesn't have any of the requested Components.
    """
    entd = world._entities[entity_id]
    return {cc: entd[cc] for cc in components}


def all_components(
    entity_id: "EntityID", world: "_World" = _default_world
) -> "Mapping[Type[Component], Component]":
    """
    Get a mapping of all Components for a specific entity in a given World.

    :param entity_id: ID of the entity to get the Component from.
    :type entity_id: `EntityID`

    :param world: The World to look up the entity in.
    :type world: Optional[`World`]

    :return: An immutable mapping between each Component type of which the
             entity has an instance, and the instance.
    :rtype: Mapping[Type[Component], Component]

    :raises KeyError: If the entity doesn't exist in this `World`.
    """
    return _MappingProxy(world._entities[entity_id])


def has_component(
    entity_id: "EntityID",
    component_type: "Type[Component]",
    world: "_World" = _default_world,
) -> "bool":
    """
    Check if a given entity has a specific Component.

    ::

        if not has_component(entity_id, FreezeStatus):
            do_stuff(entity_id)

    :param entity_id: ID of the entity to check
    :type entity_id: `EntityID`

    :param component_type: Component type to check for
    :type component_type: Type[:class:`~snecs.Component`]

    :param world: World to check for the entity
    :type world: Optional[`World`]

    :return: Whether the entity has a component of the given type.
    :rtype: bool

    :raises KeyError: If the entity doesn't exist in this `World`.
    """
    return component_type in world._entities[entity_id]


def has_components(
    entity_id: "EntityID",
    component_types: "Iterable[Type[Component]]",
    world: "_World" = _default_world,
) -> "bool":
    """
    Check if a given entity has all of the specified Components.

    :param entity_id: ID of the entity to check
    :type entity_id: `EntityID`

    :param component_types: Iterable of component types to check for
    :type component_types: Iterable[Type[:class:`~snecs.Component`]]

    :param world: World to check for the entity
    :type world: Optional[`World`]

    :return: Whether the entity has components of all of the given types.
    :rtype: bool

    :raises KeyError: If the entity doesn't exist in this `World`.
    """
    d = world._entities[entity_id]
    return all(c in d for c in component_types)


def remove_component(
    entity_id: "EntityID",
    component_type: "Type[Component]",
    world: "_World" = _default_world,
) -> "None":
    """
    Remove the component of a given type from an entity.

    :param entity_id: ID of the entity to remove the component from.
    :type entity_id: `EntityID`

    :param component_type: Type of the component to remove.
    :type component_type: Type[:class:`~snecs.Component`]

    :param world: The World to look up the entity in.
    :type world: Optional[`World`]

    :raises KeyError: If the entity doesn't exist in this `World` or it
                      doesn't have the specified component.
    """
    del world._entities[entity_id][component_type]
    world._entity_cache[component_type].remove(entity_id)
    world._entity_bitmasks[entity_id] ^= component_type._bitmask


def schedule_for_deletion(
    entity_id: "EntityID", world: "_World" = _default_world
) -> "None":
    """
    Schedule an entity for deletion from the World. Thread-safe.

    The entity will be deleted on the next call to `process_pending_deletions`.

    This method is idempotent - you can schedule an entity for deletion many
    times, and it will only be deleted once.

    .. warning::

        This will **not** raise an error if the entity doesn't exist.
        However, calling this with an entity ID that isn't in the World
        will put the World into an unrecoverable state - all later calls to
        `process_pending_deletions` will fail.

        To prevent this from happening, make sure you're calling
        `process_pending_deletions` only *after* all your Systems have
        finished running for this frame, so that they won't have a reference
        to the dead entity on the next loop.

        Calling `delete_entity_immediately` on an entity that is scheduled
        for deletion but hasn't been deleted yet is safe, and will remove the
        entity from the deletion queue.

        If there is a risk of an entity you want to delete being already
        gone, do this instead::

            from snecs.ecs import exists, schedule_for_deletion

            if exists(entity_id, world):
                schedule_for_deletion(entity_id, world)


    :param entity_id: ID of the entity to schedule for deletion
    :type entity_id: `EntityID`

    :param world: The World to delete the entity from
    :type world: Optional[`World`]
    """
    world._entities_to_delete.add(entity_id)


def exists(entity_id: "EntityID", world: "_World") -> "bool":
    """
    Check whether an entity exists in this World.

    This is a convenience function that should not be necessary except in
    concurrent code or with uncertain usage of `schedule_for_deletion` (see
    the documentation of that function).

    :param entity_id: ID of the entity to check for
    :type entity_id: `EntityID`

    :param world: The World to check for the existence of the entity
    :type world: `World`

    :return: True if the entity exists in the World, False otherwise
    :rtype: bool
    """
    return entity_id in world._entities


def delete_entity_immediately(
    entity_id: "EntityID", world: "_World" = _default_world
) -> "None":
    """
    Delete an entity from a given `World` immediately. *Not* thread-safe.

    Idempotent. Will *not* throw a KeyError if the entity doesn't exist in
    this `World`.

    :param entity_id: ID of the entity to delete
    :type entity_id: `EntityID`

    :param world: The World to delete the entity from
    :type world: Optional[`World`]
    """
    ctypes = world._entities.pop(entity_id, ())
    for ct in ctypes:
        world._entity_cache[ct].discard(entity_id)
    world._entity_bitmasks.pop(entity_id, None)
    world._entities_to_delete.discard(entity_id)


def process_pending_deletions(world: "_World" = _default_world) -> "None":
    """
    Process pending entity deletions.

    Equivalent to calling `delete_entity_immediately` on all
    entities for which `schedule_for_deletion` had been called for since
    the last call to ``process_pending_deletions``.

    Idempotent. *Not* thread-safe.

    :param world: The World to delete the entities from
    :type world: Optional[`World`]
    """
    # Make a copy so that we can safely iterate over it
    for entid in list(world._entities_to_delete):
        delete_entity_immediately(entid, world)


#: See `serialize_world`
SERIALIZED_COMPONENTS_KEY = "C"
#: See `serialize_world`
SERIALIZED_ENTITIES_KEY = "E"


# This, sadly, has to have extremely loose typing until at least PyPy hits 3.8.
# TypedDict came to `typing` way too late.
# There's lots of type: ignore here, thanks to that.
def serialize_world(world: "_World" = _default_world) -> "SerializedWorldType":
    """
    Serialize a World and all the Entities and Components inside it.

    **All Component classes in your World must be serializable for this to
    run.**

    Returns a Python dictionary that can be passed to `deserialize_world` to
    reconstruct the serialized world.

    You can for example, dump the output of `serialize_world` to a file with
    `json.dump`, to use as an effortless savegame format::

        import json

        def save_to_file(world: World, filename: str):
            with open(filename, "w") as f:
                json.dump(serialize_world(world), f)

    The resulting dict looks like this::

        serialized = {
            SERIALIZED_COMPONENTS_KEY: ["ComponentType1",...],
            SERIALIZED_ENTITIES_KEY: {
                entity_id: {
                    component_idx: serialized_component_data,
                    ...
                },
                entity_id: ...,
                ...
            }
        }

    Where:

    - ``serialized[SERIALIZED_COMPONENTS_KEY]`` is a list of names of the
      Component classes present in the World.
    - ``serialized[SERIALIZED_ENTITIES_KEY]`` is a dictionary mapping entity
      IDs to dictionaries representing their components.

      - Each of the dictionaries in ``serialized[SERIALIZED_ENTITIES_KEY]``
        maps indices of component types in
        ``serialized[SERIALIZED_COMPONENTS_KEY]`` to that component's
        serialized data.

        For example, if ``serialized[SERIALIZED_COMPONENTS_KEY]`` is
        ``[A, B, C]``, the index ``1`` would indicate that the value is a
        serialized instance of a ``B`` component.

    :param world: The World to serialize.
    :type world: Optional[`World`]

    :return: A serialized dictionary representing this World, which can be
             deserialized with `deserialize_world`.
    :rtype: `SerializedWorldType`

    :raise AttributeError: If any component in the World is not serializable.
    """
    # A mapping of component type names to indices in `component_types`
    reverse_components = {}
    component_types: "List[str]" = []

    entities = {}
    for ent_id, components in world._entities.items():
        serialized: "Dict[int, Any]" = {}  # type: ignore
        for ctype, cmpnt in components.items():
            if ctype not in reverse_components:
                idx = reverse_components[ctype] = len(component_types)
                component_types.append(ctype.__name__)
            else:
                idx = reverse_components[ctype]
            serialized[idx] = cmpnt.serialize()  # type: ignore
        entities[ent_id] = serialized  # type: ignore

    return {  # type: ignore
        SERIALIZED_COMPONENTS_KEY: component_types,
        SERIALIZED_ENTITIES_KEY: entities,  # type: ignore
    }


def deserialize_world(
    serialized: "SerializedWorldType", name: "Optional[str]" = None
) -> "_World":
    """
    Deserialize the output of `serialize_world` into a new `World`.

    If any of the component types that were registered in the serialized World
    are not registered when you call `deserialize_world`, or any of them
    have been renamed, this will fail and raise a ValueError.

    :param serialized: A serialized world, as output by `serialize_world`.
    :type serialized: `SerializedWorldType`

    :param name: An optional name to use for the new World. See `World`.
    :type name: Optional[str]

    :return: A new World with the data from the serialized one.
    :rtype: `World`
    """
    world = _World(name=name)
    serialized_names = _cast(
        "List[str]", serialized[SERIALIZED_COMPONENTS_KEY]
    )
    component_types: "List[Type[Component]]" = [
        _component_names[name] for name in serialized_names
    ]
    serialized_entities = _cast(
        "Dict[EntityID, Dict[int, str]]", serialized[SERIALIZED_ENTITIES_KEY]
    )
    for ent_id, components in serialized_entities.items():
        ent_id = int(ent_id)
        cmp_instances = [
            component_types[int(i)].deserialize(serialized)
            for i, serialized in components.items()
        ]
        if ent_id > world._entity_counter:
            world._entity_counter = ent_id
        # This is pretty directly copied over from `add_component`
        bitmask = _ZERO
        entdict = {}
        entcache = world._entity_cache
        for component in cmp_instances:
            cc = component.__class__
            bitmask |= component._bitmask
            entdict[cc] = component
            entcache.setdefault(cc, set()).add(ent_id)
        world._entities[ent_id] = entdict
        world._entity_bitmasks[ent_id] = bitmask

    return world



# Copyright (c) 2020 Slavfox
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Non-public implementation details for internal snecs use.

If you thought “that just sounds like an euphemism for ‘dirty hacks’” to
yourself, well... you're absolutely correct.

The names defined here are not part of the public API, and subject to change.
"""
from typing import TYPE_CHECKING, Any, Mapping, TypeVar, Union
from abc import ABC, abstractmethod

__all__ = ["InvariantDict", "Bitmask", "ZERO"]


if TYPE_CHECKING:
    from typing import Iterable, NoReturn, Iterator, Optional

    class _Dict:
        """Dict, but ignored by mypy."""

        # Here's the problem:
        # I want InvariantDict to keep the native dict performance, so it has
        # to inherit from dict. At the same time, calling most normal dict
        # methods on an InvariantDict is not valid, and I'd like the type
        # system to reflect this.
        # Ideally, I'd like to mark InvariantDict as a subtype of Mapping,
        # but not Dict.
        # This is why, for Mypy, we make `_Dict` is a no-op class, but for the
        # interpreter, it's literally just the builtin `dict.


else:
    _Dict = dict

_IntOrBitmask = Union[int, "Bitmask"]


class Bitmask(int):
    """
    Effectively a ``NewType(int)`` that allows bitwise operations.
    """

    __slots__ = ()

    # Normally, typing.NewType would be perfect for Bitmasks - after all,
    # we never want to pass a random integer into a function expecting a
    # Bitmask. Aliasing `Bitmask = int` doesn't help us at all.
    # However, NewType has a big problem::
    #
    #     foo: Bitmask = Bitmask(0b10)
    #     foo |= 0b01  # error: Incompatible types in assignment (expression
    #                  # has type "int", variable has type "Bitmask")
    #     foo = Bitmask(foo | 0b01)  # ok
    #
    # Why is this a problem? Because explicitly casting to Bitmask is slooooow.
    # I'm not kidding. The last line is twice as slow as the previous one on
    # my machine. That's the last thing we want when we're already using
    # bitmasks for performance!
    #
    # So, here's where the ugly hack comes in. We let Mypy know that
    # `Bitmask(0b10) | 0b01` is still a valid Bitmask, but at runtime, this
    # entire class body is just::
    #
    #     class Bitmask(int):
    #         pass
    #
    # That way, it doesn't screech at us when we use bitwise operators,
    # but at the same time, it raises warnings when we try to use a Bitmask
    # improperly - eg. by adding to, or multiplying it - or when we pass a
    # naked `int` into a function expecting a Bitmask.
    #
    # Yay.
    if TYPE_CHECKING:

        def __lshift__(self, other: "int") -> "Bitmask":
            ...

        def __rshift__(self, other: "int") -> "Bitmask":
            ...

        def __and__(self, other: "_IntOrBitmask") -> "Bitmask":
            ...

        def __xor__(self, other: "_IntOrBitmask") -> "Bitmask":
            ...

        def __or__(self, other: "_IntOrBitmask") -> "Bitmask":
            ...

        def __invert__(self) -> "Bitmask":
            ...


ZERO = Bitmask(0)


K = TypeVar("K")
V = TypeVar("V")


class InvariantDict(_Dict, Mapping[K, V], ABC):
    """
    A mapping with auto-assigned keys that rejects mutation other than add().
    """

    __slots__ = ()

    @abstractmethod
    def value_for(self, key: "K") -> "V":
        ...

    def add(self, key: "K") -> "V":
        v = self.value_for(key)
        dict.__setitem__(self, key, v)  # type: ignore
        return v

    def __setitem__(self, key: "K", value: "V") -> "NoReturn":
        raise TypeError(
            f"{self.__class__.__name__} does not support item assignment, "
            f"use .add() instead."
        )

    def update(  # type: ignore
        self, *args: "Any", **kwargs: "Any"
    ) -> "NoReturn":
        raise TypeError(
            f"{self.__class__.__name__} does not support .update(), use "
            f".add() instead."
        )

    def __delitem__(self, key: "K") -> "NoReturn":
        raise TypeError(
            f"{self.__class__.__name__} does not support key removal."
        )

    def setdefault(
        self, key: "K", default: "Optional[V]" = None
    ) -> "NoReturn":
        raise TypeError(
            f"{self.__class__.__name__} does not support .setdefault(), use "
            f".get() instead."
        )

    def pop(self, key: "K", default: "Optional[V]" = None) -> "NoReturn":
        raise TypeError(f"{self.__class__.__name__} does not support .pop().")

    def popitem(self) -> "NoReturn":
        raise TypeError(
            f"{self.__class__.__name__} does not support .popitem(). Use "
            f"reversed() instead."
        )

    @classmethod
    def fromkeys(
        cls, iterable: "Iterable[K]", value: "Optional[V]" = None
    ) -> "InvariantDict[K, V]":
        if value is not None:
            raise TypeError(
                f"{cls.__name__} does not support two-argument fromkeys()."
            )
        d = cls()
        for k in iterable:
            d.add(k)
        return d

    # You already know how this works.
    if TYPE_CHECKING:

        def __getitem__(self, item: "K") -> "V":
            ...

        def __iter__(self) -> "Iterator[K]":
            ...

        def __len__(self) -> "int":
            ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['snecs']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'snecs',
    'version': '1.1.1',
    'description': 'A straightforward, nimble ECS for Python',
    'long_description': '.. image:: https://raw.githubusercontent.com/slavfox/snecs/master/docs/_static/snecs_logo.png\n   :align: center\n   :alt: snecs\n\n=====\nsnecs\n=====\n\nA straightforward, nimble ECS for Python.\n\n.. teaser-start\n\n|PyPI badge| |PyVersion badge| |PyImplementation badge| |Mypy badge| |License badge|\n\n``snecs`` is a pure Python 3.6+, dependency-free\n`ECS <https://snecs.slavfox.space/ecs/>`__ library,\nheavily inspired by Rust’s\n`Legion <https://github.com/TomGillen/legion>`__, and aiming to be as\nfast and easy-to-use as possible.\n\n.. |PyPI badge| image:: https://img.shields.io/pypi/v/snecs\n   :alt: PyPI\n   :target: https://pypi.org/project/snecs/\n\n.. |PyVersion badge| image:: https://img.shields.io/pypi/pyversions/snecs\n   :alt: PyPI - Python Version\n\n.. |PyImplementation badge| image:: https://img.shields.io/pypi/implementation/snecs\n   :alt: PyPI - Implementation\n\n.. |Mypy badge| image:: https://img.shields.io/badge/mypy-typed-informational\n   :alt: Mypy: checked\n   :target: http://mypy-lang.org/\n\n.. |License badge| image:: https://img.shields.io/github/license/slavfox/snecs\n   :alt: GitHub\n   :target: https://github.com/slavfox/snecs/blob/master/LICENSE\n\n.. teaser-end\n\nScroll down to learn more, or `check out the documentation!\n<https://snecs.slavfox.space>`_\n\nOverview\n========\n\n``snecs`` is an\n`ECS <https://en.wikipedia.org/wiki/Entity_component_system>`__ library,\nwritten from the ground up to be:\n\nStraightforward!\n----------------\n\nThe ``snecs`` API is designed to be both easy-to-use, and encourage cleanly\nstructured code. It follows a simple principle - functions do things,\nclasses represent things - in an attempt to reduce the incidence of\nantipatterns like\n`ClassesAsNamespaces <https://www.youtube.com/watch?v=o9pEzgHorH0>`__.\n\nNimble!\n-------\n\n``snecs`` is written with a benchmark-driven approach. Every statement in\nthe hot path is benchmarked against alternative ways to express the same\nbehavior, to let your code run as fast as you need it to.\n\nOne of the design goals is outrunning\n`esper <https://github.com/benmoran56/esper>`__, and eventually I’ll\nhave a benchmark suite to post here.\n\nDependency-free!\n----------------\n\n``snecs`` has `no dependencies\nwhatsoever <https://github.com/slavfox/snecs/blob/master/pyproject.toml>`__,\nother than Python 3.6 or higher; you won’t need to worry about\ndeprecation warnings from dependencies, or having to install systemwide\nlibraries for dependencies to work - because there are none! Installing\nsnecs is as simple as running:\n\n.. code:: console\n\n   $ pip install snecs\n\nPure Python!\n------------\n\n``snecs`` is written entirely in Python. It does not use any modules written\nin C or any other language, which means that you don’t have to worry\nabout needing a C compiler if there are no wheels for your platform and\ndistribution, and that it **works perfectly under PyPy**, and gets the\nfull benefit of its JIT compiler and optimizations.\n\nFully typed!\n------------\n\n``snecs`` is checked against a `very aggressive mypy\nconfiguration <https://github.com/slavfox/snecs/blob/master/mypy.ini>`__\nto catch bugs and let you fully enjoy the library even when writing\ntype-annotated code. You won’t have to ``# type: ignore`` any use of\nsnecs.\n\nLovingly-commented!\n-------------------\n\nOver a third of the non-blank lines in the ``snecs/`` directory are\ncomments. If you ever need to dive into the code to see how ``snecs`` works\nunder the hood, they will guide you through the source and explain every\nworkaround, optimization, and obscure trick you might find non-obvious.\n\nLicense\n=======\n\n``snecs`` is made available under the terms of the Mozilla Public License\nVersion 2.0, the full text of which is available\n`here <https://www.mozilla.org/en-US/MPL/2.0/>`__, and included in\n`LICENSE <https://github.com/slavfox/snecs/blob/master/LICENSE>`__. If\nyou have questions about the license, check Mozilla’s `MPL\nFAQ <https://www.mozilla.org/en-US/MPL/2.0/FAQ/>`__.\n',
    'author': 'Slavfox',
    'author_email': 'slavfoxman@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/slavfox/snecs',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
